/*
 *   This file is part of PKSM-Core
 *   Copyright (C) 2016-2020 Bernardo Giordano, Admiral Fish, piepie62
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   Additional Terms 7.b and 7.c of GPLv3 apply to this file:
 *       * Requiring preservation of specified reasonable legal notices or
 *         author attributions in that material or in the Appropriate Legal
 *         Notices displayed by works containing it.
 *       * Prohibiting misrepresentation of the origin of that material,
 *         or requiring that modified versions of such material be marked in
 *         reasonable ways as different from the original version.
 */

#include "sav/SavBW.hpp"
#include "utils/crypto.hpp"
#include "utils/endian.hpp"
#include <algorithm>

namespace pksm
{
    SavBW::SavBW(std::shared_ptr<u8[]> dt) : Sav5(dt, 0x80000)
    {
        game = Game::BW;

        PCLayout             = 0x0;
        Trainer1             = 0x19400;
        Trainer2             = 0x21200;
        BattleSubway         = 0x21D00;
        Party                = 0x18E00;
        PokeDex              = 0x21600;
        PokeDexLanguageFlags = 0x21920;
        WondercardFlags      = 0x1C800;
        WondercardData       = 0x1C900;
        PouchHeldItem        = 0x18400;
        PouchKeyItem         = 0x188D8;
        PouchTMHM            = 0x18A24;
        PouchMedicine        = 0x18BD8;
        PouchBerry           = 0x18C98;
        Box                  = 0x400;
    }

    void SavBW::resign(void)
    {
        const u8 blockCount = 70;

        for (u8 i = 0; i < blockCount; i++)
        {
            u16 cs = pksm::crypto::ccitt16(&data[blockOfs[i]], lengths[i]);
            LittleEndian::convertFrom<u16>(&data[chkMirror[i]], cs);
            LittleEndian::convertFrom<u16>(&data[chkofs[i]], cs);
        }
    }

    std::map<Sav::Pouch, std::vector<int>> SavBW::validItems() const
    {
        return {
            {Pouch::NormalItem,
                {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 55, 56, 57, 58, 59, 60, 61,
                    62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81,
                    82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100,
                    101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 116, 117, 118, 119,
                    135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 213, 214,
                    215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230,
                    231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246,
                    247, 248, 249, 250, 251, 252, 253, 254, 255, 256, 257, 258, 259, 260, 261, 262,
                    263, 264, 265, 266, 267, 268, 269, 270, 271, 272, 273, 274, 275, 276, 277, 278,
                    279, 280, 281, 282, 283, 284, 285, 286, 287, 288, 289, 290, 291, 292, 293, 294,
                    295, 296, 297, 298, 299, 300, 301, 302, 303, 304, 305, 306, 307, 308, 309, 310,
                    311, 312, 313, 314, 315, 316, 317, 318, 319, 320, 321, 322, 323, 324, 325, 326,
                    327, 492, 493, 494, 495, 496, 497, 498, 499, 500, 537, 538, 539, 540, 541, 542,
                    543, 544, 545, 546, 547, 548, 549, 550, 551, 552, 553, 554, 555, 556, 557, 558,
                    559, 560, 561, 562, 563, 564, 571, 572, 573, 575, 576, 577, 580, 581, 582, 583,
                    584, 585, 586, 587, 588, 589, 590}},
            {Pouch::KeyItem, {437, 442, 447, 450, 465, 466, 471, 504, 533, 574, 578, 579, 616, 617,
                                 621, 623, 624, 625, 626}},
            {Pouch::TM,
                {328, 329, 330, 331, 332, 333, 334, 335, 336, 337, 338, 339, 340, 341, 342, 343,
                    344, 345, 346, 347, 348, 349, 350, 351, 352, 353, 354, 355, 356, 357, 358, 359,
                    360, 361, 362, 363, 364, 365, 366, 367, 368, 369, 370, 371, 372, 373, 374, 375,
                    376, 377, 378, 379, 380, 381, 382, 383, 384, 385, 386, 387, 388, 389, 390, 391,
                    392, 393, 394, 395, 396, 397, 398, 399, 400, 401, 402, 403, 404, 405, 406, 407,
                    408, 409, 410, 411, 412, 413, 414, 415, 416, 417, 418, 419, 618, 619, 620, 420,
                    421, 422, 423, 424, 425}},
            {Pouch::Medicine, {17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33,
                                  34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
                                  50, 51, 52, 53, 54, 134, 504, 565, 566, 567, 568, 569, 570, 591}},
            {Pouch::Berry, {149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162,
                               163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176,
                               177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190,
                               191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204,
                               205, 206, 207, 208, 209, 210, 211, 212}}};
    }
}
